/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.TagSupport;


/**
 * This is an abstract implementation of the {@link NodeDiscovery} interface and is useful to {@link NodeDiscovery}
 * implementations that are also a Tag.  This abstract class simplifies the work required when an implementation of the
 * NodeDiscovery will also provide services as a tag and provides an attribute to set the title.
 * <p/>
 * This abstract class will implement {@link #doStartTag()}, call {@link #getCurrentNode(PageContext)} to get the
 * current node and then set the result as a page scope attribute using {@link BreadCrumbTrailTag
 * #BREADCRUMB_CURRENT_NODE}.
 * <p/>
 * <p/>
 * It is strongly recommended that implementations of {@link NodeDiscovery}, who also serve as {@link
 * javax.servlet.jsp.tagext.Tag}s, inherit from this class.  The subclass then does not need to worry about tag
 * rendering
 * <p/>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 */
public abstract class AbstractNodeDiscovery extends TagSupport implements NodeDiscovery {

    /**
     * This field represents the title associated with this node.  This variable is set initially to <tt>null</tt>, is
     * mutable and can be <tt>null</tt> but never an empty string.  A <tt>null</tt> title means that the title for the
     * current node should default that found in the {@link PathDiscovery} data source. This variable is referenced by
     * the associated getter/setter methods
     */
    private String title = null;

    /**
     * Constructs this abstract base class, simply an empty constructor used as a placeholder for future logic.
     */
    protected AbstractNodeDiscovery() {
        // this constructor is empty intentionally
    }

    /**
     * This method will get the current node url and if found, save it to {@link BreadCrumbTrailTag
     * #BREADCRUMB_CURRENT_NODE} attribute using page scope.
     *
     * @return Always {@link #SKIP_BODY}
     *
     * @throws JspException if an exception occurred (none should however)
     */
    public int doStartTag() throws JspException {
        final BreadCrumbNode crumbNode = getCurrentNode(pageContext);
        if (crumbNode != null) {
            pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_CURRENT_NODE, crumbNode, PageContext.PAGE_SCOPE);
        }
        return SKIP_BODY;
    }

    /**
     * This method returns the value of the title property.
     *
     * @return The possibly <tt>null</tt>, non-empty string representing the title
     */
    public String getTitle() {
        return title;
    }

    /**
     * This method sets the value of the title property.
     *
     * @param title A possibly <tt>null</tt>, possibly empty (trim'd) string representing the title
     */
    public void setTitle(final String title) {
        if (title != null && title.trim().length() == 0) {
            this.title = null;
        } else {
            this.title = title;
        }
    }
}
