/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail;

import java.io.Serializable;


/**
 * A BreadCrumbNode is a structure for holding url and the page title of node.
 * <p/>
 * <b>Thread Safety:</b> This class is immutable and thread safe.
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 1.0
 */
public class BreadCrumbNode implements Serializable {

    /**
     * This field represents the title associated with this node.  This field is set by the constructor and is
     * immutable.  Its value can be <tt>null</tt> if the title should default to the title found in the {@link
     * PathDiscovery}. This value should not be <tt>null</tt> if it's a node that is part of a path returned by the
     * {@link PathDiscovery} implementation.  This variable is referenced by the associated getter.
     *
     * @since 1.0
     */
    private final String title;

    /**
     * This field represents the url (including query string) associated with this node.  This variable is set by the
     * constructor, is immutable and will never be <tt>null</tt> or empty String.  This variable is referenced by the
     * associated getter.
     *
     * @since 1.0
     */
    private final String url;

    /**
     * Constructs a new BreadCrumbNode from the specified arguments.
     *
     * @param url   A non-<tt>null</tt>, non-empty (trim'd) url of page represented by this node
     * @param title A possibly <tt>null</tt>, possibly empty (trim'd) title of page represented by this node
     *
     * @throws IllegalArgumentException if url is <tt>null</tt> or an empty (trim'd) string
     * @since 2.0
     */
    public BreadCrumbNode(final String url, final String title) {
        if (url == null) {
            throw new IllegalArgumentException("The parameter named [url] was null.");
        }
        if (url.trim().length() == 0) {
            throw new IllegalArgumentException("The parameter named [url] was an empty String.");
        }

        this.url = url;
        if (title != null && title.trim().length() == 0) {
            this.title = null;
        } else {
            this.title = title;
        }
    }

    /**
     * This method returns the title of the node.
     *
     * @return A possibly <tt>null</tt>, non-empty (trim'd) string representing the title of the node
     *
     * @since 1.0
     */
    public String getTitle() {
        return title;
    }

    /**
     * This method returns the Url of the node.  The returned string will be the URL with any query string.
     *
     * @return A non-<tt>null</tt>, non-empty (trim'd) string representing the url (including query string) of the node
     *
     * @since 1.0
     */
    public String getUrl() {
        return url;
    }
}
