/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail;

import com.topcoder.util.config.ConfigManager;
import com.topcoder.util.config.UnknownNamespaceException;
import com.topcoder.util.objectfactory.InvalidClassSpecificationException;
import com.topcoder.util.objectfactory.ObjectFactory;
import com.topcoder.util.objectfactory.impl.ConfigManagerSpecificationFactory;
import com.topcoder.util.objectfactory.impl.IllegalReferenceException;
import com.topcoder.util.objectfactory.impl.SpecificationConfigurationException;
import com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.XmlPathDiscovery;

import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.BodyTagSupport;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.ListIterator;


/**
 * This is the main tag class that applications will use. This tag will render the current node to the root node (or
 * vice-versa) using a path separator and specific styles. The application can setup the separator and the various
 * styles either by specifying default values in the configuration manager or by specifying attributes on the string.
 * The object factory will be used to discover the node and the path to the root. The node discovery is optional. If the
 * node discovery is not specified, it is assumed a node discovery tag was used on the page to define the current node.
 * The body of the tag will be used to format the various trails (using variables defined by this class).
 * <p/>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 1.0
 */
public class BreadCrumbTrailTag extends BodyTagSupport {

    /**
     * This static constant represents the page scope attribute name that the current node, if found, would be found
     * under.  This constant will be used by a {@link NodeDiscovery} instance to store the current node in the attribute
     * map and by this class to retrieve.
     *
     * @since 2.0
     */
    public static final String BREADCRUMB_CURRENT_NODE = "breadCrumbCurrentNode";

    /**
     * This static constant represents the nested variable name that the current node to format would be found under.
     * This constant will be used by this class to define a variable available to a NodeFormatter to format. Note: this
     * is the page scope attribute name that will be made available as a variable via the {@link
     * BreadCrumbTrailTagExtraInfo} class.
     *
     * @since 2.0
     */
    public static final String BREADCRUMB_FORMAT_NODE = "breadCrumbFormatNode";

    /**
     * <p>This static constant represents the nested variable name that the current node's style that will be used to
     * format.  This constant will be used by this class to define a variable available to a NodeFormatter to format.
     * Note: this is the page scope attribute name that will be made available as a variable via the {@link
     * BreadCrumbTrailTagExtraInfo} class.
     *
     * @since 2.0
     */
    public static final String BREADCRUMB_FORMAT_STYLE_CLASS = "breadCrumbFormatStyleClass";

    /**
     * This static constant represents a "pathDirection" constant for formatting the path from the root (on the left) to
     * the current node (on the right).  This variable is only referenced in the {@link #doStartTag()} but is public for
     * a convenience to strategy implementations.
     *
     * @since 1.0
     */
    public static final int ROOT_TO_CURRENT = 0;

    /**
     * This static constant represents a "pathDirection" constant for formatting the path from the current node (on the
     * left) to the root node (on the right).  This variable is only referenced in the {@link #doStartTag()} but is
     * public for a convenience to strategy implementations.
     *
     * @since 1.0
     */
    public static final int CURRENT_TO_ROOT = 1;

    /**
     * This is the {@link ConfigManager} namespace used for configuring the tag defaults.
     *
     * @since 2.0
     */
    private static final String NAMESPACE = "com.topcoder.web.ui.tag.breadcrumbtrail";

    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the path separator
     * property is looked up.
     *
     * @see #pathSeparator
     * @since 2.0
     */
    private static final String PATH_SEPARATOR_PROPERTY_NAME = "pathSeparator";
    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the path separator style
     * property is looked up.
     *
     * @see #pathSeparatorStyle
     * @since 2.0
     */
    private static final String PATH_SEPARATOR_STYLE_PROPERTY_NAME = "pathSeparatorStyle";
    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the node style property
     * is looked up.
     *
     * @see #nodeStyle
     * @since 2.0
     */
    private static final String NODE_STYLE_PROPERTY_NAME = "nodeStyle";
    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the current node style
     * property is looked up.
     *
     * @see #currentNodeStyle
     * @since 2.0
     */
    private static final String CURRENT_NODE_STYLE_PROPERTY_NAME = "currentNodeStyle";
    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the root node style
     * property is looked up.
     *
     * @see #rootNodeStyle
     * @since 2.0
     */
    private static final String ROOT_NODE_STYLE_PROPERTY_NAME = "rootNodeStyle";
    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the mouse over style
     * property is looked up.
     *
     * @see #mouseOverStyle
     * @since 2.0
     */
    private static final String MOUSE_OVER_STYLE_PROPERTY_NAME = "mouseOverStyle";
    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the silent errors
     * property is looked up.
     *
     * @see #silentErrors
     * @since 2.0
     */
    private static final String SILENT_ERRORS_PROPERTY_NAME = "silentErrors";
    /**
     * This is the name of the {@link ConfigManager} property from which the default value for the path direction
     * property is looked up.
     *
     * @see #pathDirection
     * @since 2.0
     */
    private static final String PATH_DIRECTION_PROPERTY_NAME = "pathDirection";
    /**
     * This field represents the separator that will be rendered between nodes in a path. This variable is set by the
     * constructor, is mutable and will never be <tt>null</tt> but can be an empty (trim'd) string. This variable is
     * referenced by the associated {@link #getPathSeparator()}/{@link #setPathSeparator(String)} methods and the {@link
     * #doAfterBody()} method.
     *
     * @since 1.0
     */
    private String pathSeparator;

    /**
     * This field represents the separator CSS style that will be used to draw the separator.  This is the actual CSS
     * style that will be rendered to the underlying text. This variable is set by the constructor, is mutable and can
     * be <tt>null</tt> or an empty (trim'd) string to specify no specific style (i.e. use the style that is currently
     * in context). This variable is referenced by the associated {@link #getPathSeparatorStyle()}/{@link
     * #setPathSeparator(String)} methods and the {@link #doAfterBody()} method.
     *
     * @since 1.0
     */
    private String pathSeparatorStyle;

    /**
     * This field represents the node (non-root and non-current) CSS style that will be used to draw the node's title.
     * This is the actual CSS style that will be rendered to the underlying link. This variable is set by the
     * constructor, is mutable and can be <tt>null</tt> or an empty (trim'd) string to specify no specific style (i.e.
     * use the style that is currently in context). This variable is referenced by the associated {@link
     * #getNodeStyle()}/{@link #setNodeStyle(String)} methods and the {@link #doAfterBody()} method.
     *
     * @since 1.0
     */
    private String nodeStyle;

    /**
     * This field represents the current (only) node CSS style that will be used to draw the node's title.  This is the
     * actual CSS style that will be rendered to the underlying link. This variable is set by the constructor, is
     * mutable and can be <tt>null</tt> or an empty (trim'd) string to specify no specific style (i.e. use the style
     * that is currently in context). This variable is referenced by the associated {@link
     * #getCurrentNodeStyle()}/{@link #setCurrentNodeStyle(String)} methods and the {@link #doAfterBody()} method.
     *
     * @since 1.0
     */
    private String currentNodeStyle;

    /**
     * This field represents the root (only) node CSS style that will be used to draw the node's title.  This is the
     * actual CSS style that will be rendered to the underlying link. This variable is set by the constructor, is
     * mutable and can be <tt>null</tt> or an empty (trim'd) string to specify no specific style (i.e. use the style
     * that is currently in context). This variable is referenced by the associated {@link #getRootNodeStyle()}/{@link
     * #setRootNodeStyle(String)} methods and the {@link #doAfterBody()} method.
     *
     * @since 1.0
     */
    private String rootNodeStyle;

    /**
     * This field represents the mouse over node CSS style that will be used to draw the node's title.  This is the
     * actual CSS style that will be rendered to the underlying link. This variable is set by the constructor, is
     * mutable and can be <tt>null</tt> or an empty (trim'd) string to specify no specific style (i.e. use the style
     * that is currently in context). This variable is referenced by the associated {@link #getMouseOverStyle()}/{@link
     * #setMouseOverStyle(String)} methods and the {@link #doAfterBody()} method.
     *
     * @since 1.0
     */
    private String mouseOverStyle;

    /**
     * This field represents whether any exceptions should be silently ignore or not (<tt>true</tt> to silently ignore,
     * <tt>false</tt> to let exception pass through to JSP engine). This variable is set by the constructor and is
     * mutable. This variable is referenced by the associated {@link #getSilentErrors()}/{@link
     * #setSilentErrors(boolean)} methods and the {@link #doStartTag()}, {@link #doAfterBody()}and {@link #doEndTag()}
     * methods.
     *
     * @since 1.0
     */
    private boolean silentErrors;

    /**
     * This field represents the direction of the path.  The current allowable values are either {@link
     * #ROOT_TO_CURRENT} (specifying root [on the left] to current [on the right]) or {@link #CURRENT_TO_ROOT}
     * (specifying current [on the left] to root [on the right]). This variable is set by the constructor, is mutable
     * and can only be the above values. This variable is referenced by the associated {@link
     * #getPathDirection()}/{@link #setPathDirection(int)} methods and the {@link #doStartTag()} method.
     *
     * @since 1.0
     */
    private int pathDirection;

    /**
     * This field represents the a path override for this specific tag. This variable is set initially to <tt>null</tt>
     * and is mutable.  If <tt>null</tt>, there is no path override available and normal path discovery should be used.
     * If non-<tt>null</tt>, a XmlPathDiscovery class will be constructed with this path and used to return the path
     * specified in this variable. Please see {@link XmlPathDiscovery} class for details on it's formatting. This
     * variable is referenced by the associated {@link #getPathOverride()}/{@link #setPathOverride(String)} methods and
     * the {@link #doStartTag()} method.
     *
     * @since 2.0
     */
    private String pathOverride = null;

    /**
     * This field represents title that will be used if the node is being discovered by this tag (i.e. no other tag has
     * already discovered the node and set the {@link #BREADCRUMB_CURRENT_NODE} attribute) - otherwise it is ignored.
     * This variable is set initially to <tt>null</tt>, is mutable and never will be an empty string but can be
     * <tt>null</tt>.  A <tt>null</tt> indicates that the title should default to the title specified in the
     * PathDiscovery data source.  This variable is referenced by the associated {@link #getTitle()}/{@link
     * #setTitle(String)} methods and the {@link #doStartTag()} method.
     *
     * @since 2.0
     */
    private String title = null;

    /**
     * This field represents the iterator over the current node path.  This iterator will initially be <tt>null</tt>, be
     * mutable by the {@link #doStartTag()} and will be non-<tt>null</tt> during the body processing.  The iterator will
     * be created in the {@link #doStartTag()} from the path discovered by the {@link PathDiscovery} implementation. The
     * node formatter (or body if no node formatter specified) will then be called, for each node in this iterator.
     *
     * @since 2.0
     */
    private ListIterator iterator = null;

    /**
     * Constructs the BreadCrumbTrailTag using the configuration manager to set default values.
     *
     * @since 1.0
     */
    public BreadCrumbTrailTag() {
        // the setters are called instead of setting fields directly
        // to ensure the validity of field values
        setPathSeparator(lookupOptionalStringFromConfigManager(NAMESPACE, PATH_SEPARATOR_PROPERTY_NAME, ":"));
        setPathSeparatorStyle(lookupOptionalStringFromConfigManager(NAMESPACE, PATH_SEPARATOR_STYLE_PROPERTY_NAME,
            null));
        setNodeStyle(lookupOptionalStringFromConfigManager(NAMESPACE, NODE_STYLE_PROPERTY_NAME, null));
        setCurrentNodeStyle(lookupOptionalStringFromConfigManager(NAMESPACE, CURRENT_NODE_STYLE_PROPERTY_NAME, null));
        setRootNodeStyle(lookupOptionalStringFromConfigManager(NAMESPACE, ROOT_NODE_STYLE_PROPERTY_NAME, null));
        setMouseOverStyle(lookupOptionalStringFromConfigManager(NAMESPACE, MOUSE_OVER_STYLE_PROPERTY_NAME, null));
        silentErrors = Boolean.valueOf(lookupOptionalStringFromConfigManager(NAMESPACE, SILENT_ERRORS_PROPERTY_NAME,
            "false"))
            .booleanValue();
        final String pathDirectionString = lookupOptionalStringFromConfigManager(NAMESPACE,
            PATH_DIRECTION_PROPERTY_NAME, "0");
        int direction = ROOT_TO_CURRENT;
        try {
            direction = Integer.parseInt(pathDirectionString);
        } catch (NumberFormatException e) {
            // this parse error is ignored silently and the default value is used instead
        }
        if (direction != ROOT_TO_CURRENT && direction != CURRENT_TO_ROOT) {
            direction = ROOT_TO_CURRENT;
        }
        setPathDirection(direction);
    }

    /**
     * This method returns the value of the pathDirection property.
     *
     * @return The path direction - either {@link #ROOT_TO_CURRENT} or {@link #CURRENT_TO_ROOT}
     *
     * @since 1.0
     */
    public int getPathDirection() {
        return pathDirection;
    }

    /**
     * This method sets the value of the pathDirection property.
     *
     * @param pathDirection The new path direction
     *
     * @throws IllegalArgumentException if given pathDirection is not {@link #ROOT_TO_CURRENT} or {@link
     *                                  #CURRENT_TO_ROOT}
     * @since 1.0
     */
    public void setPathDirection(final int pathDirection) {
        if (pathDirection != ROOT_TO_CURRENT && pathDirection != CURRENT_TO_ROOT) {
            throw new IllegalArgumentException(
                "The parameter named [pathDirection] was expected to be either ROOT_TO_CURRENT [" + ROOT_TO_CURRENT
                    + "] or CURRENT_TO_ROOT [" + CURRENT_TO_ROOT + "] , but was [" + pathDirection + "].");
        }
        this.pathDirection = pathDirection;
    }

    /**
     * This method returns the value of the pathSeparator property.
     *
     * @return A non-<tt>null</tt>, possibly empty (trim'd) string representing the path separator
     *
     * @since 1.0
     */
    public String getPathSeparator() {
        return pathSeparator;
    }

    /**
     * This method sets the value of the pathSeparator property.
     *
     * @param pathSeparator A non-<tt>null</tt>, possibly empty (trim'd) string representing the path separator
     *
     * @throws IllegalArgumentException if the pathSeparator is <tt>null</tt>
     * @since 1.0
     */
    public void setPathSeparator(final String pathSeparator) {
        if (pathSeparator == null) {
            throw new IllegalArgumentException("The parameter named [pathSeparator] was null.");
        }

        this.pathSeparator = pathSeparator;
    }

    /**
     * This method returns the value of the pathSeparatorStyle Property.
     *
     * @return A possibly <tt>null</tt>, never empty (trim'd) string representing the path separator style
     *
     * @since 1.0
     */
    public String getPathSeparatorStyle() {
        return pathSeparatorStyle;
    }

    /**
     * This method sets the value of the pathSeparatorStyle property.
     *
     * @param pathSeparatorStyle A possibly <tt>null</tt>, possibly empty (trim'd) string representing the path
     *                           separator style (empty strings will be translated to a <tt>null</tt>)
     *
     * @since 1.0
     */
    public void setPathSeparatorStyle(final String pathSeparatorStyle) {
        this.pathSeparatorStyle = simplifyEmptyStringToNull(pathSeparatorStyle);
    }

    /**
     * This method returns the value of the nodeStyle property.
     *
     * @return A possibly <tt>null</tt>, never empty (trim'd) string representing the node style
     *
     * @since 1.0
     */
    public String getNodeStyle() {
        return nodeStyle;
    }

    /**
     * This method sets the value of the nodeStyle property.
     *
     * @param nodeStyle A possibly <tt>null</tt>, possibly empty (trim'd) string representing the node style (an empty
     *                  string will be translated to a <tt>null</tt>)
     *
     * @since 1.0
     */
    public void setNodeStyle(final String nodeStyle) {
        this.nodeStyle = simplifyEmptyStringToNull(nodeStyle);
    }

    /**
     * This method returns the value of the currentNodeStyle property.
     *
     * @return A possibly <tt>null</tt>, never empty (trim'd) string representing the current node style
     *
     * @since 1.0
     */
    public String getCurrentNodeStyle() {
        return currentNodeStyle;
    }

    /**
     * This method sets the value of the currentNodeStyle property.
     *
     * @param currentNodeStyle A possibly <tt>null</tt>, possibly empty (trim'd) string representing the current node
     *                         style (an empty string will be translated to a <tt>null</tt>)
     *
     * @since 1.0
     */
    public void setCurrentNodeStyle(final String currentNodeStyle) {
        this.currentNodeStyle = simplifyEmptyStringToNull(currentNodeStyle);
    }

    /**
     * This method returns the value of the rootNodeStyle property.
     *
     * @return A possibly <tt>null</tt>, never empty (trim'd) string representing the root node style
     *
     * @since 1.0
     */
    public String getRootNodeStyle() {
        return rootNodeStyle;
    }

    /**
     * This method sets the value of the rootNodeStyle property.
     *
     * @param rootNodeStyle A possibly <tt>null</tt>, possibly empty (trim'd) string representing the root node style
     *                      (an empty string will be translated to a <tt>null</tt>)
     *
     * @since 1.0
     */
    public void setRootNodeStyle(final String rootNodeStyle) {
        this.rootNodeStyle = simplifyEmptyStringToNull(rootNodeStyle);
    }

    /**
     * This method returns the value of the mouseOverStyle property.
     *
     * @return A possibly <tt>null</tt>, never empty (trim'd) string representing the mouse over style
     *
     * @since 1.0
     */
    public String getMouseOverStyle() {
        return mouseOverStyle;
    }

    /**
     * This method sets the value of the mouseOverStyle property.
     *
     * @param mouseOverStyle A possibly <tt>null</tt>, possibly empty (trim'd) string representing the mouse over style
     *                       (an empty string will be translated to a <tt>null</tt>)
     *
     * @since 1.0
     */
    public void setMouseOverStyle(final String mouseOverStyle) {
        this.mouseOverStyle = simplifyEmptyStringToNull(mouseOverStyle);
    }

    /**
     * This method returns the value of the pathOverride property.
     *
     * @return A possibly <tt>null</tt>, never empty (trim'd) string representing the path override
     *
     * @since 2.0
     */
    public String getPathOverride() {
        return pathOverride;
    }

    /**
     * This method sets the value of the pathOverride property.
     *
     * @param pathOverride A possibly <tt>null</tt>, possibly empty (trim'd) string representing the path override (an
     *                     empty string will be translated to a <tt>null</tt>)
     *
     * @since 2.0
     */
    public void setPathOverride(final String pathOverride) {
        this.pathOverride = simplifyEmptyStringToNull(pathOverride);
    }

    /**
     * This method returns the value of the silentErrors property.
     *
     * @return <tt>true</tt> if exceptions will be silently ignored, <tt>false</tt> if they should be thrown to the JSP
     *         Engine
     *
     * @since 1.0
     */
    public boolean getSilentErrors() {
        return silentErrors;
    }

    /**
     * This method sets the value of the silentErrors property.
     *
     * @param silentErrors <tt>true</tt> if exceptions will be silently ignored, <tt>false</tt> if they should be thrown
     *                     to the JSP Engine
     *
     * @since 1.0
     */
    public void setSilentErrors(final boolean silentErrors) {
        this.silentErrors = silentErrors;
    }

    /**
     * This method returns the value of the title property.
     *
     * @return A possibly <tt>null</tt>, never empty (trim'd) string representing the current node title
     *
     * @since 2.0
     */
    public String getTitle() {
        return title;
    }

    /**
     * This method sets the value of the title property.
     *
     * @param title A possibly <tt>null</tt>, possibly empty (trim'd) string representing the title of the current node
     *              (an empty string will be translated to a <tt>null</tt>)
     *
     * @since 2.0
     */
    public void setTitle(final String title) {
        this.title = simplifyEmptyStringToNull(title);
    }

    /**
     * This method starts the rendering of the bread crumb trail tag.
     *
     * @return An integer representing what processing to do after this method executes
     *
     * @throws BreadCrumbException if creating of bread crumb trail failed and silentErrors is <tt>false</tt>
     * @since 1.0
     */
    public int doStartTag() throws BreadCrumbException {
        // retrieve the ObjectFactory to be used
        final ObjectFactory objectFactory;
        try {
            objectFactory = new ObjectFactory(new ConfigManagerSpecificationFactory(getClass().getName()));
        } catch (SpecificationConfigurationException e) {
            return throwBreadCrumbExceptionOrSkip("Error while creating ObjectFactory.", e);
        } catch (IllegalReferenceException e) {
            return throwBreadCrumbExceptionOrSkip("Error while creating ObjectFactory.", e);
        }

        final BreadCrumbNode currentNode;
        try {
            currentNode = findCurrentNode(objectFactory);
        } catch (BreadCrumbException e) {
            return throwBreadCrumbExceptionOrSkip("Error discovering the current node.", e);
        }

        if (currentNode == null) {
            // in case the node discovery is not capable of
            // determining the node, we cannot continue
            return SKIP_BODY;
        }

        // find the path for the current node
        final PathDiscovery pathDiscovery;
        if (getPathOverride() != null) {
            pathDiscovery = new XmlPathDiscovery(getPathOverride(), false);
        } else {
            try {
                pathDiscovery = (PathDiscovery) objectFactory.createObject(PathDiscovery.class);
            } catch (InvalidClassSpecificationException e) {
                return throwBreadCrumbExceptionOrSkip("Error while creating PathDiscovery from ObjectFactory.", e);
            }
        }

        //discover path
        final BreadCrumbNode[] currentPath;
        try {
            currentPath = pathDiscovery.getPath(currentNode, pageContext);
        } catch (BreadCrumbException e) {
            return throwBreadCrumbExceptionOrSkip("Error while determining current path.", e);
        }

        if (currentPath == null || currentPath.length == 0) {
            // in case the path discovery has
            // failed, we cannot continue
            return SKIP_BODY;
        }

        final List list = Arrays.asList(currentPath);
        iterator = pathDirection == ROOT_TO_CURRENT ? list.listIterator() : list.listIterator(list.size());

        if (nodeStyle != null
            || currentNodeStyle != null
            || rootNodeStyle != null
            || mouseOverStyle != null
            || pathSeparatorStyle != null) {
            final JspWriter out = pageContext.getOut();
            try {
                out.write("<span>");
                printStyleSettings(out);
            } catch (IOException e) {
                return throwBreadCrumbExceptionOrSkip("Error while rendering the breadcrumb.", e);
            }
        }
        // when we have come here without error we can start iterating the body
        return EVAL_BODY_BUFFERED;
    }

    /**
     * This initializes the first item from the iterator and sets the attribute information for what needs to be
     * formatted. It is called once before the first body evaluation.
     *
     * @since 2.0
     */
    public void doInitBody() {
        proceedIteration();
    }

    /**
     * <p>This method examines the body content text written by the tag body and writes it out and starts again in case
     * more path elements are to be rendered.
     *
     * @return either {@link #SKIP_BODY} to stop processing the body or {@link #EVAL_BODY_AGAIN} to process the next
     *         item
     *
     * @throws BreadCrumbException in case the rendering fails and silentErrors is <tt>false</tt>
     * @since 2.0
     */
    public int doAfterBody() throws BreadCrumbException {
        final JspWriter out = getPreviousOut();
        final String stringContent = bodyContent.getString();
        if (stringContent != null && stringContent.trim().length() > 0) {
            try {
                bodyContent.writeOut(out);
            } catch (IOException e) {
                return throwBreadCrumbExceptionOrSkip("Error while writing out body content.", e);
            }

        }
        bodyContent.clearBody();

        final boolean moreAvailable = pathDirection == ROOT_TO_CURRENT
                                      ? iterator.hasNext()
                                      : iterator.hasPrevious();

        // in case there are more path elements
        // available, evaluate the body again,
        // otherwise end iteration

        if (moreAvailable) {
            try {
                printSeparator(out);
            } catch (IOException e) {
                return throwBreadCrumbExceptionOrSkip("Error printing out separator.", e);
            }
            proceedIteration();
            return EVAL_BODY_AGAIN;
        } else {
            return SKIP_BODY;
        }
    }

    /**
     * This method prints the closing tag after iteration has finished.
     *
     * @return Always {@link #EVAL_PAGE}
     *
     * @throws BreadCrumbException if creating of bread crumb trail failed and silentErrors is false
     * @since 2.0
     */
    public int doEndTag() throws BreadCrumbException {
        if (nodeStyle != null
            || currentNodeStyle != null
            || rootNodeStyle != null
            || mouseOverStyle != null
            || pathSeparatorStyle != null) {
            final JspWriter out = pageContext.getOut();
            try {
                out.write("</span>");
            } catch (IOException e) {
                return throwBreadCrumbExceptionOrSkip("Error writing out end tag.", e);
            }
        }
        return EVAL_PAGE;
    }

    /**
     * This method retrieves the next item from the iterator and sets the attribute information for what needs to be
     * formatted. It is called once before the first body evaluation from {@link #doInitBody()} and then subsequently
     * after each body iteration (from {@link #doAfterBody()}) in case there is some path left to be rendered.
     *
     * @since 2.0
     */
    private void proceedIteration() {
        // depending on whether we iterate forward
        // or backward we need to determine the
        // root and current state of a node either
        // before or after the actual iteration
        boolean isRootNode = false;
        boolean isCurentNode = false;
        if (pathDirection == CURRENT_TO_ROOT) {
            isCurentNode = !iterator.hasNext();
        } else {
            isRootNode = !iterator.hasPrevious();
        }
        final BreadCrumbNode item = (BreadCrumbNode) (pathDirection == ROOT_TO_CURRENT
                                                      ? iterator.next()
                                                      : iterator.previous());
        if (pathDirection == ROOT_TO_CURRENT) {
            isCurentNode = !iterator.hasNext();
        } else {
            isRootNode = !iterator.hasPrevious();
        }
        pageContext.setAttribute(BREADCRUMB_FORMAT_NODE, item, PageContext.PAGE_SCOPE);
        // we include the matching style class name in case
        // the style class or its :hover pseudoclass are defined
        // in the enclosing span element that was output by doStartTag
        if (isRootNode && (getRootNodeStyle() != null || getMouseOverStyle() != null)) {
            pageContext.setAttribute(BREADCRUMB_FORMAT_STYLE_CLASS, "root-node-style", PageContext.PAGE_SCOPE);
        } else if (isCurentNode && (getCurrentNodeStyle() != null || getMouseOverStyle() != null)) {
            pageContext.setAttribute(BREADCRUMB_FORMAT_STYLE_CLASS, "current-node-style", PageContext.PAGE_SCOPE);
        } else if (getNodeStyle() != null || getMouseOverStyle() != null) {
            pageContext.setAttribute(BREADCRUMB_FORMAT_STYLE_CLASS, "node-style", PageContext.PAGE_SCOPE);
        } else {
            // clear node class in case none applies
            // otherwise we have the wrong value from previous iteration
            pageContext.removeAttribute(BREADCRUMB_FORMAT_STYLE_CLASS);
        }
    }

    /**
     * This method prints the style settings out to the specified writer.
     *
     * @param out The non-<tt>null</tt> JspWriter to write the style settings to
     *
     * @throws IllegalArgumentException if the out is <tt>null</tt>
     * @throws IOException              if an io occurs writing out the style
     * @since 2.0
     */
    private void printStyleSettings(final JspWriter out) throws IOException {
        if (out == null) {
            throw new IllegalArgumentException("The parameter named [out] was null.");
        }

        out.write("<style>");
        if (getPathSeparatorStyle() != null) {
            printStyle(out, ".path-separator-style", getPathSeparatorStyle());
        }
        if (getNodeStyle() != null) {
            printStyle(out, ".node-style", getNodeStyle());
        }
        if (getCurrentNodeStyle() != null) {
            printStyle(out, ".current-node-style", getCurrentNodeStyle());
        }
        if (getRootNodeStyle() != null) {
            printStyle(out, ".root-node-style", getRootNodeStyle());
        }
        if (getMouseOverStyle() != null) {
            printStyle(out, ".node-style:hover", getMouseOverStyle());
            printStyle(out, ".root-node-style:hover", getMouseOverStyle());
            printStyle(out, ".current-node-style:hover", getMouseOverStyle());
        }
        out.write("</style>\n");
    }

    /**
     * This method prints the specified style out to the specified writer.
     *
     * @param out   The non-<tt>null</tt> JspWriter to write the style settings to
     * @param style The non-<tt>null</tt>, non-empty (trim'd) string representing the style to write
     * @param value The non-<tt>null</tt>, non-empty (trim'd) string representing the value of the style to write
     *
     * @throws IllegalArgumentException if any argument is <tt>null</tt> or any string argument is an empty (trim'd)
     *                                  string
     * @throws IOException              if an io occurs writing out the style
     * @since 2.0
     */
    private void printStyle(final JspWriter out, final String style, final String value) throws IOException {
        if (out == null) {
            throw new IllegalArgumentException("The parameter named [out] was null.");
        }
        if (style == null) {
            throw new IllegalArgumentException("The parameter named [style] was null.");
        }
        if (value == null) {
            throw new IllegalArgumentException("The parameter named [value] was null.");
        }
        if (style.trim().length() == 0) {
            throw new IllegalArgumentException("The parameter named [style] was an empty String.");
        }
        if (value.trim().length() == 0) {
            throw new IllegalArgumentException("The parameter named [value] was an empty String.");
        }

        out.write(style);
        out.write(" {");
        out.write(value);
        out.write("}\n");
    }

    /**
     * This method prints the separator (with the separator style) to the specified writer.
     *
     * @param out The non-<tt>null</tt> JspWriter to write the style settings to
     *
     * @throws IllegalArgumentException if the out is <tt>null</tt>
     * @throws IOException              if an io occurs writing out the style
     * @since 2.0
     */
    private void printSeparator(final JspWriter out) throws IOException {
        if (out == null) {
            throw new IllegalArgumentException("The parameter named [out] was null.");
        }

        if (getPathSeparatorStyle() != null) {
            out.write("<span class=\"path-separator-style\">");
        }
        out.write(getPathSeparator());
        if (getPathSeparatorStyle() != null) {
            out.write("</span>");
        }
    }

    /**
     * This method handles a caught exception. Depending on whether the silentErrors property is <tt>true</tt> or false,
     * the method either throws a BreadCrumbException with the specified message and cause or simply returns {@link
     * #SKIP_BODY} to signal that the current JSP rendering process shall be aborted.
     *
     * @param message the message to be used in case an exception is thrown
     * @param ex      the exception that was caught and needs to be used as cause for the thrown exception
     *
     * @return {@link #SKIP_BODY} in case silentErrors is <tt>true</tt>
     *
     * @throws BreadCrumbException created from the given arguments if silentErrors is <tt>false</tt>
     * @since 2.0
     */
    private int throwBreadCrumbExceptionOrSkip(final String message, final Throwable ex) throws
        BreadCrumbException {
        if (silentErrors) {
            return SKIP_BODY;
        } else {
            throw new BreadCrumbException(message, ex);
        }
    }

    /**
     * This method retrieves the current node from the page context. If that fails, the given ObjectFactory is used to
     * construct a {@link NodeDiscovery} instance from which the currentNode is discovered.
     *
     * @param objectFactory the ObjectFactory to be used for {@link NodeDiscovery} construction
     *
     * @return the current node discovered, or <tt>null</tt> if discovery failed
     *
     * @throws BreadCrumbException in case the node discovery fails
     */
    private BreadCrumbNode findCurrentNode(ObjectFactory objectFactory) throws BreadCrumbException {
        // find current node either because it is available in page context or by explicitly discovering it
        final Object nodeObj = pageContext.getAttribute(BREADCRUMB_CURRENT_NODE);
        if (nodeObj instanceof BreadCrumbNode) {
            return (BreadCrumbNode) nodeObj;
        }
        // do the default discovery as current node is still unknown
        final NodeDiscovery discovery;

        try {
            discovery = (NodeDiscovery) objectFactory.createObject(NodeDiscovery.class);
        } catch (InvalidClassSpecificationException e) {
            throw new BreadCrumbException("Error while creating NodeDiscovery from ObjectFactory.", e);
        }

        if (discovery instanceof AbstractNodeDiscovery && getTitle() != null) {
            ((AbstractNodeDiscovery) discovery).setTitle(getTitle());
        }
        return discovery.getCurrentNode(pageContext);
    }

    /**
     * This method looks up an optional configuration property of type {@link String} from the {@link
     * com.topcoder.util.config.ConfigManager}.
     *
     * @param namespace    A non-<tt>null</tt>, non-empty (trim'd) string representing the {@link
     *                     com.topcoder.util.config.ConfigManager} namespace in which to lookup the configuration
     *                     property
     * @param key          A non-<tt>null</tt>, non-empty (trim'd) string representing the name of the property to be
     *                     looked up
     * @param defaultValue the value to be returned in case the value could not be found in config manager or was
     *                     <tt>null</tt> in config manager
     *
     * @return the String value looked up from the {@link com.topcoder.util.config.ConfigManager} or the given
     *         defaultValue if the value cannot be found, or was <tt>null</tt>.
     *
     * @throws IllegalArgumentException if namespace or key are <tt>null</tt> or empty (trim'd) strings
     * @since 2.0
     */
    private static String lookupOptionalStringFromConfigManager(final String namespace, final String key,
                                                                final String defaultValue) {
        if (namespace == null) {
            throw new IllegalArgumentException("The parameter named [namespace] was null.");
        }
        if (key == null) {
            throw new IllegalArgumentException("The parameter named [key] was null.");
        }
        if (namespace.trim().length() == 0) {
            throw new IllegalArgumentException("The parameter named [namespace] was an empty String.");
        }
        if (key.trim().length() == 0) {
            throw new IllegalArgumentException("The parameter named [key] was an empty String.");
        }
        try {

            final String lookedUp = ConfigManager.getInstance().getString(namespace, key);
            if (lookedUp == null) {
                return defaultValue;
            }
            return lookedUp;
        } catch (UnknownNamespaceException e) {
            return defaultValue;
        }
    }

    /**
     * This method converts empty (trim'd) Strings into <tt>null</tt>s.
     *
     * @param string the String to be converted
     *
     * @return <tt>null</tt> if the given String was null or empty (trim'd) String, the given String otherwise
     *
     * @since 2.0
     */
    private static String simplifyEmptyStringToNull(final String string) {
        if (string != null && string.trim().length() == 0) {
            return null;
        }
        return string;
    }

}
