/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import com.topcoder.util.objectfactory.InvalidClassSpecificationException;
import com.topcoder.util.objectfactory.ObjectFactory;
import com.topcoder.util.objectfactory.impl.ConfigManagerSpecificationFactory;
import com.topcoder.util.objectfactory.impl.IllegalReferenceException;
import com.topcoder.util.objectfactory.impl.SpecificationConfigurationException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.PathDiscovery;

import javax.servlet.jsp.PageContext;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * This implementation of  {@link PathDiscovery} will attempt to create a static site map and then apply dynamic
 * information to that site map. The dynamic information can either be in matching nodes with different query strings or
 * in new nodes that will be assumed to map to the last known node.
 * <p/>
 * <p/>
 * <b>Thread Safety:</b> This class is immutable and thus thread safe. But as it accesses state that it writes and reads
  * to the page context multiple instances working on the same sessions could produce unexpected results.
  *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 */
public class DynamicPathDiscovery implements PathDiscovery {

    /**
     * This static constant represents the session scope attribute name that will contain the current known path. This
     * constant will be used by this class to retrieve the current path. The current path will be of <tt>
     * NodePair[]</tt> type representing the know path from the root (0th index) to the prior page (lenght-1th index).
     */
    public static final String CURRENT_PATH = DynamicPathDiscovery.class.getName() + ".currentPath";

    /**
     * This static constant represents the session scope attribute name that will contain the site map. This constant
     * will be used by this class to retrieve the site map.  The site map will be a {@link Map} with {@link NodeMatcher}
     * as the key and a {@link List} of {@link NodeMatcher}s as the value and can be empty.  This map represents the
     * {@link NodeMatcher} parent to it's children {@link NodeMatcher}(s) if any.  The list can be empty if there are no
     * children of the parent (i.e. a leaf node).
     */
    public static final String CURRENT_SITEMAP = DynamicPathDiscovery.class.getName() + ".siteMap";
    /**
     * This static constant represents the session scope attribute name that will contain the reverse site map. This
     * constant will be used by this class to retrieve the site map.  The site map will be a {@link Map} with {@link
     * NodeMatcher} as the key and a {@link List} of {@link NodeMatcher}s as the value and can be empty.  This map
     * represents the {@link NodeMatcher} child to it's parents {@link NodeMatcher}(s) if any.  The list can be empty if
     * there are no parents of the child (i.e. a root node).
     */
    public static final String CURRENT_REVERSE_SITEMAP = DynamicPathDiscovery.class.getName() + ".reverseSiteMap";

    /**
     * Constructs the DynamicPathDiscovery object.
     */
    public DynamicPathDiscovery() {
        // This constructor is intentionally empty
    }

    /**
     * This method attempts to determine the page from the current node to the root and return that path.
     *
     * @param currentNode A non-<tt>null</tt> current node
     * @param pageContext A non-<tt>null</tt> page context
     *
     * @return <tt>null</tt> if the path can't be found or a non-empty array representing the path from root node (0th
     *         index) to current node (size-1th index)
     *
     * @throws IllegalArgumentException if either arg is <tt>null</tt>
     * @throws BreadCrumbException      if an exception occurred while processing
     */
    public BreadCrumbNode[] getPath(final BreadCrumbNode currentNode, final PageContext pageContext)
        throws BreadCrumbException {
        if (currentNode == null) {
            throw new IllegalArgumentException("The parameter named [currentNode] was null.");
        }
        if (pageContext == null) {
            throw new IllegalArgumentException("The parameter named [pageContext] was null.");
        }

        NodePair[] currentPath = (NodePair[]) pageContext.getAttribute(CURRENT_PATH, PageContext.SESSION_SCOPE);

        final Map siteMap;
        final Object sessionSiteMap = pageContext.getAttribute(CURRENT_SITEMAP, PageContext.SESSION_SCOPE);
        if (sessionSiteMap instanceof Map) {
            siteMap = (Map) sessionSiteMap;
        } else {
            // invalidate current path if using new site map
            currentPath = null;
            siteMap = createSiteMapFromObjectFactory();
        }

        final Map reverseMap;
        final Object sessionReverseMap = pageContext.getAttribute(CURRENT_REVERSE_SITEMAP, PageContext.SESSION_SCOPE);
        if (sessionReverseMap instanceof Map) {
            reverseMap = (Map) sessionReverseMap;
        } else {
            reverseMap = DiscoveryUtils.createReverseMap(siteMap);
        }

        NodePair[] foundPath = null;

        if (currentPath != null) {
            // check whether node is on current path
            foundPath = checkOnCurrentPath(currentPath, currentNode);

            // we do not need to check for siteMap!=null,
            // as a null siteMap would have made currentPath=null
            if (foundPath == null) {
                // check whether the node is one-off any of current path elements
                foundPath = checkOneOffCurrentPath(currentPath, siteMap, currentNode);
            }
        }

        if (foundPath == null && siteMap != null) {

            foundPath = findShortestPathInSiteMap(siteMap, currentNode, reverseMap, currentPath);
        }

        // we now 'discover' the new path by appending the current node to the current path
        if (foundPath == null && currentPath != null) {
            // we append the node to the current path
            foundPath = new NodePair[currentPath.length + 1];
            System.arraycopy(currentPath, 0, foundPath, 0, currentPath.length);
            foundPath[currentPath.length] = new NodePair(currentNode, null);
        }

        // save back current state
        pageContext.setAttribute(CURRENT_PATH, foundPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(CURRENT_REVERSE_SITEMAP, reverseMap, PageContext.SESSION_SCOPE);

        if (foundPath == null || foundPath.length == 0) {
            return null;
        } else {
            // copy the path from the foundPath
            final BreadCrumbNode[] ret = new BreadCrumbNode[foundPath.length];
            for (int i = 0; i < foundPath.length; i++) {
                ret[i] = foundPath[i].getNode();
            }
            return ret;
        }
    }

    /**
     * This method creates a new  site map by creating a {@link SiteDiscovery} instance and calling its {@link
     * SiteDiscovery#getSiteMap()} method.
     *
     * @return the site map created
     *
     * @throws BreadCrumbException in case the creation fails
     */
    private Map createSiteMapFromObjectFactory() throws BreadCrumbException {
        final ObjectFactory objectFactory;
        try {
            objectFactory = new ObjectFactory(new ConfigManagerSpecificationFactory(getClass().getName()));
        } catch (SpecificationConfigurationException e) {
            throw new BreadCrumbException("Error creating ObjectFactory.", e);
        } catch (IllegalReferenceException e) {
            throw new BreadCrumbException("Error creating ObjectFactory.", e);
        }
        try {
            final SiteDiscovery siteDiscovery = (SiteDiscovery) objectFactory.createObject(SiteDiscovery.class);
            return siteDiscovery.getSiteMap();
        } catch (InvalidClassSpecificationException e) {
            throw new BreadCrumbException("Error while creating SiteDiscovery via ObjectFactory.", e);
        }
    }

    /**
     * This method does a search over all elements of the site map and in case a match has been found in the site map
     * the shortest (i.e. least steps to a root node)  and most common (most elements in common with current path) path
     * is returned. In case such a path is found, it is returned. Otherwise <tt>null</tt> is returned.
     *
     * @param siteMap     the site map searched for matches with the given node
     * @param currentNode the node to find a path for
     * @param reverseMap  the reverse map to determine the path to a root node
     * @param currentPath the current path to use as commonality comparison argument
     *
     * @return the found path for the given node, or <tt>null</tt> if none has been found
     *
     * @throws IllegalArgumentException if any arg except currentPath is <tt>null</tt>
     */
    private NodePair[] findShortestPathInSiteMap(final Map siteMap, final BreadCrumbNode currentNode,
                                                 final Map reverseMap,
                                                 final NodePair[] currentPath) {
        if (siteMap == null) {
            throw new IllegalArgumentException("The parameter named [siteMap] was null.");
        }
        if (currentNode == null) {
            throw new IllegalArgumentException("The parameter named [currentNode] was null.");
        }
        if (reverseMap == null) {
            throw new IllegalArgumentException("The parameter named [reverseMap] was null.");
        }

        // find matching node matchers in the site map
        final List matching = new ArrayList();
        final Set candidates = siteMap.keySet();
        for (Iterator iterator = candidates.iterator(); iterator.hasNext();) {
            final NodeMatcher nodeMatcher = (NodeMatcher) iterator.next();
            if (nodeMatcher.matches(currentNode)) {
                matching.add(nodeMatcher);
            }
        }

        if (!matching.isEmpty()) {
            //we have a matching matcher and need a shortest-path-decision

            final List shorthestPaths = DiscoveryUtils.findShortestPaths(matching, reverseMap);
            if (shorthestPaths == null || shorthestPaths.isEmpty()) {
                return null;
            }
            List shorthestPath = null;
            int maxFoundCommonalityCount = -1;
            if (currentPath != null) {
                // this algorithm takes the most-common (i.e. the one
                // that has the most in common with currentPath) of
                // the found shortest paths
                for (Iterator iterator = shorthestPaths.iterator(); iterator.hasNext();) {
                    final List shortestPathCandidate = (List) iterator.next();
                    int idx = 0;
                    for (Iterator iterator1 = shortestPathCandidate.iterator();
                         iterator1.hasNext() && idx < currentPath.length; idx++) {
                        final NodeMatcher currentNodeMatcher = currentPath[idx].getNodeMatcher();
                        if (currentNodeMatcher == null || !currentNodeMatcher.equals(iterator1.next())) {
                            break;
                        }
                    }
                    if (idx > maxFoundCommonalityCount) {
                        // this one is better than the best found one, so store it
                        shorthestPath = shortestPathCandidate;
                        maxFoundCommonalityCount = idx;
                    }
                }

            } else {
                // we have no currentPath to compare with so we take the first one
                shorthestPath = DiscoveryUtils.getFirstPath(shorthestPaths);
                //no commonality, zero elements copied (as currentPath is null anyway)
                maxFoundCommonalityCount = 0;
            }

            //build up the path from the selected shortest path
            final NodePair[] foundPath = new NodePair[shorthestPath.size()];
            // copy the elements in-common
            if (currentPath != null) {
                System.arraycopy(currentPath, 0, foundPath, 0, maxFoundCommonalityCount);
            }

            //build new elements for the uncommon part
            for (int i = maxFoundCommonalityCount; i < foundPath.length; i++) {
                final NodeMatcher pathElement = (NodeMatcher) shorthestPath.get(i);
                if (i == foundPath.length - 1) {
                    // the last path element does always contain the currentNode
                    // instead of a synthetical node made from a matcher

                    foundPath[i] = new NodePair(currentNode, pathElement);
                } else {
                    foundPath[i] = new NodePair(new BreadCrumbNode(pathElement.getUrl(), pathElement.getTitle()),
                        pathElement);
                }
            }
            return foundPath;
        }
        return null;
    }

    /**
     * This method checks whether there is a match for the given node one off the current path (i.e. one of the child
     * nodes of the last path element with a NodeMatcher does match). In case such a path is found, it is returned.
     * Otherwise <tt>null</tt> is returned.
     *
     * @param currentPath the current path
     * @param siteMap     the site map to use for lookup of node children
     * @param currentNode the node to find a path for
     *
     * @return the found path for the given node, or <tt>null</tt> if none has been found
     *
     * @throws IllegalArgumentException if any arg is <tt>null</tt>
     */
    private NodePair[] checkOneOffCurrentPath(final NodePair[] currentPath, final Map siteMap,
                                              final BreadCrumbNode currentNode) {
        for (int i = currentPath.length - 1; i >= 0; i--) {
            final NodePair pathElem = currentPath[i];
            final NodeMatcher nodeMatcher = pathElem.getNodeMatcher();
            if (nodeMatcher != null) {
                final List childPaths = (List) siteMap.get(nodeMatcher);
                if (childPaths != null) {
                    for (Iterator iterator = childPaths.iterator(); iterator.hasNext();) {
                        final NodeMatcher childPath = (NodeMatcher) iterator.next();
                        if (childPath.matches(currentNode)) {
                            final NodePair[] foundPath = new NodePair[i + 2];
                            System.arraycopy(currentPath, 0, foundPath, 0, i + 1);
                            foundPath[i + 1] = new NodePair(currentNode, childPath);
                            return foundPath;
                        }
                    }
                }
                // we do only check the last path element that has a
                // matcher, so we break after finding one that had a matcher
                break;
            }
        }
        return null;
    }

    /**
     * This method checks whether there is a sub-path of the current path that matches the given currentNode. In case
     * such a path is found, it is returned. Otherwise <tt>null</tt> is returned.
     *
     * @param currentPath the path to check for match with the given node
     * @param currentNode the node to check for match with the given path
     *
     * @return the found path for the given node, or <tt>null</tt> if none has been found
     *
     * @throws IllegalArgumentException if any arg is <tt>null</tt>
     */
    private NodePair[] checkOnCurrentPath(final NodePair[] currentPath, final BreadCrumbNode currentNode) {
        if (currentPath == null) {
            throw new IllegalArgumentException("The parameter named [currentPath] was null.");
        }
        if (currentNode == null) {
            throw new IllegalArgumentException("The parameter named [currentNode] was null.");
        }

        for (int i = currentPath.length - 1; i >= 0; i--) {
            final NodePair pathElem = currentPath[i];
            final NodeMatcher nodeMatcher = pathElem.getNodeMatcher();
            if ((nodeMatcher == null && (pathElem.getNode().getUrl().equals(currentNode.getUrl())))
                || (nodeMatcher != null && nodeMatcher.matches(currentNode))) {
                final NodePair[] foundPath = new NodePair[i + 1];
                System.arraycopy(currentPath, 0, foundPath, 0, i);
                foundPath[i] = new NodePair(currentNode, nodeMatcher);
                return foundPath;
            }
        }
        return null;
    }

}
