/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;

import java.io.Serializable;


/**
 * This interface defines the contract for a node matcher. Implementations need to provide three functions: <ol> <li>A
 * matching function that will take a node and determine if it matches</li> <li>A default title string used for those
 * nodes that match</li> <li>A default url to use if no contextual information is available</li> </ol>
 * <p/>
 * <b>Implementation notes: </b> Implementations of this interface <b>must</b> override the {@link
 * Object#equals(Object)}/{@link Object#hashCode()} methods because the implementation will be stored in a map (likely a
 * hash map).
 * <p/>
 * {@link NodeMatcher} implementations will be created by {@link SiteDiscovery} implementations when the site map is
 * created. The {@link NodeMatcher} implementation will then be used by the {@link DynamicPathDiscovery} when it
 * attempts to discover the path to the root node from some current node.
 * <p/>
 * <b>Thread Safety:</b> Implementations of this interface should be thread safe.
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 */
public interface NodeMatcher extends Serializable {
    /**
     * This method determines whether the specified node matches this node matcher.
     *
     * @param node A non-<tt>null</tt> node to test
     *
     * @return <tt>true</tt> if the node matches, <tt>false</tt> otherwise
     *
     * @throws IllegalArgumentException if the node is not <tt>null</tt>
     */
    public boolean matches(BreadCrumbNode node);

    /**
     * This method returns the default title that nodes that match this node matcher can use if their title is not
     * specified.
     *
     * @return A non-<tt>null</tt>, non-empty (trim'd) string representing the default title
     */
    public String getTitle();

    /**
     * This method returns the default url that can be used if contextual node information is not available.
     *
     * @return A non-<tt>null</tt>, non-empty (trim'd) string representing the url
     */
    public String getUrl();
}


