/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.PathDiscovery;
import com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.nodematchers.RegexNodeMatcher;
import org.apache.xml.utils.DefaultErrorHandler;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.servlet.jsp.PageContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * This implementation of the {@link PathDiscovery} <b>and</b> {@link SiteDiscovery} will attempt to create a static
 * site map from either a filename or string source (both of which is an XML document).
 * <p/>
 * This class will parse the document to create a static map of {@link RegexNodeMatcher}s (where each node can provide a
 * regex matching). If used as a {@link PathDiscovery}, this provides a static view of the site (i.e. no dynamic
 * capabilities).
 * <p/>
 * <b>Thread Safety:</b> This class is immutable and thus thread safe. But as it accesses state that it writes and reads
 * to the page context multiple instances working on the same sessions could produce unexpected results.
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 */
public class XmlPathDiscovery implements SiteDiscovery, PathDiscovery {

    /**
     * This static constant represents the session scope attribute name that will contain the site map. This constant
     * will be used by this class to retrieve the site map. The site map will be a Map with NodeMatcher as the key and a
     * List of NodeMatcher as the value and can be empty. This map represents the NodeMatcher parent to it's children
     * NodeMatcher(s) if any. The list can be empty if there are no children of the parent (i.e. a leaf node).
     */
    public static final String CURRENT_SITEMAP = XmlPathDiscovery.class.getName() + ".siteMap";

    /**
     * This static constant represents the session scope attribute name that will contain the reverse site map. This
     * constant will be used by this class to retrieve the site map.  The site map will be a {@link Map} with {@link
     * NodeMatcher} as the key and a {@link List} of {@link NodeMatcher}s as the value and can be empty.  This map
     * represents the {@link NodeMatcher} child to it's parents {@link NodeMatcher}(s) if any.  The list can be empty if
     * there are no parents of the child (i.e. a root node).
     */
    public static final String CURRENT_REVERSE_SITEMAP = XmlPathDiscovery.class.getName() + ".reverseSiteMap";
    /**
     * This is a JAXP attribute that can be used to configure the schema validation language to be used by a JAXP
     * parser.
     */
    private static final String SCHEMA_LANGUAGE_ATTRIBUTE_NAME
        = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
    /**
     * This is a JAXP attribute that can be used to configure the location of the validation schema to be used by a JAXP
     * parser.
     */
    private static final String SCHEMA_SOURCE_ATTRIBUTE_NAME = "http://java.sun.com/xml/jaxp/properties/schemaSource";
    /**
     * This is the URI identifying the W3C schema language. It is used as value for the JAXP parser attribute {@link
     * #SCHEMA_LANGUAGE_ATTRIBUTE_NAME}.
     */
    private static final String W3C_SCHEMA = "http://www.w3.org/2001/XMLSchema";
    /**
     * This is the resource path to the XSD file containing the schema describing the bread crumb trail tag site map
     * format expected by this SiteDiscovery.
     */
    private static final String BREAD_CRUMB_TRAIL_SCHEMA_XSD_NAME = "BreadCrumbTrailSchema.xsd";

    /**
     * This field represents the source of the site data. The source will be set by the constructor, is immutable and
     * will never be <tt>null</tt> but may be an empty (trim'd) string. This variable is referencable by the {@link
     * #getSiteMap()} method.
     */
    private final String source;

    /**
     * This field represents whether the source variable represents a file (<tt>true</tt>) or a raw XML string
     * (<tt>false</tt>). The isFile will be set by the constructor and is immutable. This variable is referencable by
     * the {@link getSiteMap()} method.
     */
    private final boolean isFile;

    /**
     * Constructs the XmlPathDiscovery from the given filename.
     *
     * @param filename A non-<tt>null</tt>, non-empty (trim'd) string representing the file to use
     *
     * @throws IllegalArgumentException if the filename is a <tt>null</tt> or empty (trim'd) string
     */
    public XmlPathDiscovery(final String filename) {
        this(filename, true);
    }

    /**
     * Constructs the XmlPathDiscovery from the given source and whether that source is a direct XML string or a
     * filename.
     *
     * @param source A non-<tt>null</tt>, non-empty (trim'd) string representing the source
     * @param isFile Whether the source is a file (<tt>true</tt>) or a raw XML string (<tt>false</tt>)
     *
     * @throws IllegalArgumentException if the source is <tt>null</tt> or an empty (trim'd) string
     */
    public XmlPathDiscovery(final String source, final boolean isFile) {
        if (source == null) {
            throw new IllegalArgumentException("The parameter named [source] was null.");
        }
        if (source.trim().length() == 0) {
            throw new IllegalArgumentException("The parameter named [source] was an empty String.");
        }

        this.source = source;
        this.isFile = isFile;
    }

    /**
     * This method returns a map of {@link NodeMatcher} keys with a {@link java.util.List} of {@link NodeMatcher}
     * elements as the value. The key {@link NodeMatcher} represents the parent and the value (the list) represents the
     * child values. The value can be <tt>null</tt> or an empty list to represent no child elements (i.e. it's a leaf
     * node).
     *
     * @return A non-<tt>null</tt> map of parent {@link NodeMatcher} to a {@link java.util.List} of {@link NodeMatcher}
     *         children
     *
     * @throws BreadCrumbException if parsing of data source failed
     */
    public Map getSiteMap() throws BreadCrumbException {
        final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        final DocumentBuilder db;

        try {
            factory.setNamespaceAware(true);
            factory.setValidating(true);
            // say parser to use schema validation
            factory.setAttribute(SCHEMA_LANGUAGE_ATTRIBUTE_NAME,
                W3C_SCHEMA);
            // say parser to use the XSD from the component jar,
            // loaded via class loader from component jar
            // This is done as the XSD in the component jar is
            // the one that has been included by the developer
            // and thus is the schema describing the format
            // expected by the component code, whereas any
            // externally provided version of that XSD file
            // could be different from what the code expects
            factory.setAttribute(SCHEMA_SOURCE_ATTRIBUTE_NAME,
                getClass().getClassLoader().getResourceAsStream(BREAD_CRUMB_TRAIL_SCHEMA_XSD_NAME));
            db = factory.newDocumentBuilder();
        } catch (ParserConfigurationException pce) {
            throw new BreadCrumbException("Error while creating DOM parser.", pce);
        }
        // set an error handler on the parser,
        // as the error handler that is assigned by
        // default does only print errors to stdout,
        // as we want errors and fatal errors in the
        // XML to cause an exception during parsing
        // we use the DefaultErrorHandler
        // implementation -- not to be mistaken with
        // the DefaultHandler implementation -- which
        // does throw Exceptions on errors and fatal errors.
        db.setErrorHandler(new DefaultErrorHandler());

        try {

            final Document document;
            if (isFile) {
                final InputStream stream;

                final File file = new File(source);
                if (file.isAbsolute()) {
                    // check if source denotes absolute path
                    stream = new FileInputStream(file);
                } else {
                    // check in class loader for resource as relative path
                    stream = getClass().getClassLoader().getResourceAsStream(source);
                }
                if (stream != null) {
                    document = db.parse(stream);
                } else {
                    document = db.parse(source);
                }
            } else {
                document = db.parse(new InputSource(new StringReader(source)));
            }

            final Map ret = new HashMap();
            parseChildren(document.getDocumentElement(), ret);
            return ret;
        } catch (IOException ioe) {
            throw new BreadCrumbException("Error while parsing the input source.", ioe);
        } catch (SAXException saxe) {
            throw new BreadCrumbException("Error while parsing the input source.", saxe);
        } catch (NullPointerException npe) {
            throw new BreadCrumbException("Error while parsing the input source.", npe);
        }
    }

    /**
     * This method attempts to determine the page from the current node to the root and return that path based on the
     * current site map.
     *
     * @param currentNode A non-<tt>null</tt> current node
     * @param pageContext A non-<tt>null</tt> page context
     *
     * @return <tt>null</tt> if the path can't be found or a non-empty array representing the path from root node (0th
     *         index) to current node (size-1th index)
     *
     * @throws IllegalArgumentException if any arg is <tt>null</tt>
     * @throws BreadCrumbException      if an exception occurred while processing
     */
    public BreadCrumbNode[] getPath(final BreadCrumbNode currentNode, final PageContext pageContext)
        throws BreadCrumbException {
        if (currentNode == null) {
            throw new IllegalArgumentException("The parameter named [currentNode] was null.");
        }
        if (pageContext == null) {
            throw new IllegalArgumentException("The parameter named [pageContext] was null.");
        }

        final Map siteMap;
        final Object o = pageContext.getAttribute(CURRENT_SITEMAP, PageContext.SESSION_SCOPE);
        if (o instanceof Map) {
            siteMap = (Map) o;
        } else {
            siteMap = getSiteMap();
            pageContext.setAttribute(CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);
        }

        // find matching node matchers in the site map
        final List matching = new ArrayList();
        final Set candidates = siteMap.keySet();
        for (Iterator iterator = candidates.iterator(); iterator.hasNext();) {
            final NodeMatcher nodeMatcher = (NodeMatcher) iterator.next();
            if (nodeMatcher.matches(currentNode)) {
                matching.add(nodeMatcher);
            }
        }

        if (!matching.isEmpty()) {
            //we have a matching matcher and need a shortest-path-decision

            // the reverse map containing all parents for a given matcher,
            // look up from session or calculate from siteMap if not found
            final Map reverseMap;
            final Object sessionReverseMap = pageContext.getAttribute(CURRENT_REVERSE_SITEMAP,
                PageContext.SESSION_SCOPE);
            if (sessionReverseMap instanceof Map) {
                reverseMap = (Map) sessionReverseMap;
            } else {
                reverseMap = DiscoveryUtils.createReverseMap(siteMap);
                pageContext.setAttribute(CURRENT_REVERSE_SITEMAP, reverseMap, PageContext.SESSION_SCOPE);
            }

            final List shorthestPaths = DiscoveryUtils.findShortestPaths(matching, reverseMap);
            if (shorthestPaths == null || shorthestPaths.isEmpty()) {
                return null;
            }

            // this algorithm takes the first of the found shortest paths
            final List shorthestPath = DiscoveryUtils.getFirstPath(shorthestPaths);

            // create the breadcrumb path from the found NodeMatcher path
            final BreadCrumbNode[] ret = new BreadCrumbNode[shorthestPath.size()];
            int index = 0;
            for (Iterator iterator = shorthestPath.iterator(); iterator.hasNext(); index++) {
                final NodeMatcher currentPathElement = (NodeMatcher) iterator.next();
                if (index == ret.length - 1) {
                    //the last node shall have the currentNode's url instead of the matcher's url
                    ret[index] = updateNodeFromMatcher(currentNode, currentPathElement);
                } else {
                    ret[index] = new BreadCrumbNode(currentPathElement.getUrl(), currentPathElement.getTitle());
                }
            }

            return ret;
        } else {
            return null;
        }
    }

    /**
     * This method recursively parses part of XML document which are children of bc_node argument. All nodes are added
     * to the trailMap as RegexNodeMatcher key and have their child nodes as List of RegexNodeMatcher values.
     *
     * @param node     node for which to find children recursively
     * @param trailMap Map to which to add found nodes
     *
     * @return the list of the direct node children of the given node
     *
     * @throws IllegalArgumentException if node or trailMap is <tt>null</tt>
     * @throws BreadCrumbException      in case any invalid data is encountered during parsing
     */
    private List parseChildren(final Node node, final Map trailMap) throws BreadCrumbException {
        if (node == null) {
            throw new IllegalArgumentException("node should not be null");
        } else if (trailMap == null) {
            throw new IllegalArgumentException("trailMap should not be null");
        }

        final List ret = new ArrayList();
        final NodeList nl = node.getChildNodes();
        final int numNodes = nl.getLength();

        for (int i = 0; i < numNodes; i++) {
            final Node n = nl.item(i);

            if ("bc_node".equals(n.getNodeName())) {
                final NamedNodeMap attributes = n.getAttributes();

                final String url = getAttribute(attributes, "url");
                final String title = getAttribute(attributes, "title");
                final String pattern = getAttribute(attributes, "pattern");

                final RegexNodeMatcher currentNode = new RegexNodeMatcher(title, url, pattern);
                ret.add(currentNode);
                // this is the recursion down, parse the list of
                // children and return it as List of BreadCrumbNode
                List childList = parseChildren(n, trailMap);
                if (childList.isEmpty()) {
                    trailMap.put(currentNode, null);
                } else {
                    // if we have a non-empty child list,
                    // do check wether it is a duplicate
                    // child list definition
                    List oldChildList = (List) trailMap.get(currentNode);
                    if (oldChildList != null && !oldChildList.isEmpty()) {
                        throw new BreadCrumbException(
                            "The given site map is illegal, as it contains a child list for the node url=["
                                + currentNode.getUrl() + "] pattern=[" + currentNode.getNodePattern() + "] title=["
                                + currentNode.getTitle() + "] multiple times.");
                    }
                    trailMap.put(currentNode, childList);
                }
            }
        }
        return ret;
    }

    /**
     * This method retrieves an attribute with the given name from the given attributes map, throwing a
     * BreadCrumbException in case the attribute was not found.
     *
     * @param attributes    the attribute map to find the attribute in
     * @param attributeName the name of the attribute to be retrieved
     *
     * @return the value retrieved from the attribute map
     *
     * @throws BreadCrumbException      in case the attribute with the given name was not found in the given attribute
     *                                  map
     * @throws IllegalArgumentException if any arg is <tt>null</tt>
     */
    private static String getAttribute(final NamedNodeMap attributes, final String attributeName)
        throws BreadCrumbException {
        if (attributes == null) {
            throw new IllegalArgumentException("The parameter named [attributes] was null.");
        }
        if (attributeName == null) {
            throw new IllegalArgumentException("The parameter named [attributeName] was null.");
        }

        final Node item = attributes.getNamedItem(attributeName);
        if (item == null) {
            throw new BreadCrumbException("A bc_node element is missing its [" + attributeName + "] attribute.");
        }
        return item.getNodeValue();
    }

    /**
     * This method does return a BreadCrumbNode which has the url of the given node and either the title of the given
     * node if that was not <tt>null</tt> or the title from the given NodeMatcher.
     *
     * @param node        the node to be updated if necessary
     * @param nodeMatcher the node matcher to take the title from
     *
     * @return the possibly updated version of the given node
     *
     * @throws IllegalArgumentException if node is <tt>null</tt>
     */
    private static BreadCrumbNode updateNodeFromMatcher(final BreadCrumbNode node, final NodeMatcher nodeMatcher) {
        if (node == null) {
            throw new IllegalArgumentException("The parameter named [node] was null.");
        }
        if (nodeMatcher == null) {
            return node;
        }

        if (node.getTitle() == null) {
            return new BreadCrumbNode(node.getUrl(), nodeMatcher.getTitle());
        }
        return node;
    }
}
